@@ -1,7 +1,5 @@
 =begin
-  $Id: field.rb,v 1.11 2005/01/07 03:32:16 sam Exp $
-
-  Copyright (C) 2005 Sam Roberts
+  Copyright (C) 2008 Sam Roberts
 
   This library is free software; you can redistribute it and/or modify it
   under the same terms as the ruby language itself, see the file COPYING for
@@ -18,6 +16,17 @@ module Vpim
 
     # A field in a directory info object.
     class Field
+      # TODO
+      # - Field should know which param values and field values are
+      #   case-insensitive, configurably, so it can down case them
+      # - perhaps should have pvalue_set/del/add, perhaps case-insensitive, or
+      #   pvalue_iset/idel/iadd, where set sets them all, add adds if not present,
+      #   and del deletes any that are present
+      # - I really, really, need a case-insensitive string...
+      # - should allow nil as a field value, its not the same as '', if there is
+      #   more than one pvalue, the empty string will show up. This isn't strictly
+      #   disallowed, but its odd. Should also strip empty strings on decoding, if
+      #   I don't already.
       private_class_method :new
 
       def Field.create_array(fields)
@@ -56,8 +65,8 @@ module Vpim
 
           pvalues.each do |pvalue|
             # check if we need to do any encoding
-            if pname.downcase == 'encoding' && pvalue == :b64
-              pvalue = 'b' # the RFC definition of the base64 param value
+            if Vpim::Methods.casecmp?(pname, 'ENCODING') && pvalue == :b64
+              pvalue = 'B' # the RFC definition of the base64 param value
               value = [ value.to_str ].pack('m').gsub("\n", '')
             end
 
@@ -100,11 +109,16 @@ module Vpim
           raise Vpim::InvalidEncodingError, atline
         end
 
-        atgroup = $1
-        atname = $2
+        atgroup = $1.upcase
+        atname = $2.upcase
         paramslist = $3
         atvalue = $~[-1]
 
+        # I've seen space that shouldn't be there, as in "BEGIN:VCARD ", so
+        # strip it. I'm not absolutely sure this is allowed... it certainly
+        # breaks round-trip encoding.
+        atvalue.strip!
+
         if atgroup.length > 0
           atgroup.chomp!('.')
         else
@@ -120,24 +134,24 @@ module Vpim
           paramslist.scan( %r{#{Bnf::PARAM}}i ) do
 
             # param names are case-insensitive, and multi-valued
-            name = $1.downcase
+            name = $1.upcase
             params = $3
 
             # v2.1 params have no '=' sign, figure out what kind of param it
-            # is (either its a known encoding, or we treat it as a 'type'
+            # is (either its a known encoding, or we treat it as a 'TYPE'
             # param).
            
             if $2 == ""
               params = $1
               case $1
                 when /quoted-printable/i
-                  name = 'encoding'
+                  name = 'ENCODING'
 
                 when /base64/i
-                  name = 'encoding'
+                  name = 'ENCODING'
 
                 else
-                  name = 'type'
+                  name = 'TYPE'
               end
             end
 
@@ -148,7 +162,7 @@ module Vpim
             end
 
             params.scan( %r{#{Bnf::PVALUE}} ) do
-              atparams[name] << ($1 || $2) # Used to do this, want to stop! .downcase
+              atparams[name] << ($1 || $2)
             end
           end
         end
@@ -159,6 +173,10 @@ module Vpim
       def initialize(line) # :nodoc:
         @line = line.to_str
         @group, @name, @params, @value = Field.decode0(@line)
+
+        @params.each do |pname,pvalues|
+          pvalues.freeze
+        end
         self
       end
 
@@ -173,9 +191,9 @@ module Vpim
       # name (a String) to either a single string or symbol, or an array of
       # strings and symbols (parameters can be multi-valued).
       #
-      # If 'encoding' => :b64 is specified as a parameter, the value will be
+      # If 'ENCODING' => :b64 is specified as a parameter, the value will be
       # base-64 encoded. If it's already base-64 encoded, then use String
-      # values ('encoding' => 'b'), and no further encoding will be done by
+      # values ('ENCODING' => 'B'), and no further encoding will be done by
       # this routine.
       #
       # Currently handled value types are:
@@ -206,12 +224,13 @@ module Vpim
       # maximum line width of +width+, where +0+ means no wrapping, and nil is
       # to accept the default wrapping (75, recommended by RFC2425).
       #
-      # Note: Apple's Address Book 3.0.3 neither understands to unwrap lines
-      # when it imports vCards (it treats them as raw new-line characters), nor
-      # wraps long lines on export. On import, this is mostly a cosmetic
-      # problem, but wrapping can be disabled by setting width to +0+, if
-      # desired. On export, its a more serious bug, since Address Book wil
-      # encode the entire Note: section as a single, very long, line.
+      # Note: AddressBook.app 3.0.3 neither understands to unwrap lines when it
+      # imports vCards (it treats them as raw new-line characters), nor wraps
+      # long lines on export. This is mostly a cosmetic problem, but wrapping
+      # can be disabled by setting width to +0+, if desired.
+      #
+      # FIXME - breaks round-trip encoding, need to change this to not wrap
+      # fields that are already wrapped.
       def encode(width=nil)
         width = 75 unless width
         l = @line
@@ -236,30 +255,37 @@ module Vpim
       end
 
       # An Array of all the param names.
-      def params
+      def pnames
         @params.keys
       end
 
-      # The Array of all values of the param +name+, or nil if there is no
-      # such param.
-      #
-      # TODO - this doesn't return [] if the param isn't found, because then
-      # you can't distinguish between a param with no values, and a param that
-      # isn't defined. But maybe it isn't useful? You can always do
-      # params.include?(a_name) if you want to know if it exists. Maybe this
-      # should take a block specifying the default value, even?
-      def param(name)
-        v = @params[name.downcase]
+      # FIXME - remove my own uses of #params
+      alias params pnames # :nodoc:
+
+      # The first value of the param +name+,  nil if there is no such param,
+      # the param has no value, or the first param value is zero-length.
+      def pvalue(name)
+        v = pvalues( name )
         if v
-          v = v.collect do |p|
-            p.downcase.freeze
-          end
-          v.freeze
+          v = v.first
+        end
+        if v
+          v = nil unless v.length > 0
         end
         v
       end
 
-      alias [] param
+      # The Array of all values of the param +name+,  nil if there is no such
+      # param, [] if the param has no values. If the Field isn't frozen, the
+      # Array is mutable.
+      def pvalues(name)
+        @params[name.upcase]
+      end
+
+      # FIXME - remove my own uses of #param
+      alias param pvalues # :nodoc:
+
+      alias [] pvalues
 
       # Yield once for each param, +name+ is the parameter name, +value+ is an
       # array of the parameter values.
@@ -276,32 +302,36 @@ module Vpim
       # Note: Both the RFC 2425 encoding param ("b", meaning base-64) and the
       # vCard 2.1 encoding params ("base64", "quoted-printable", "8bit", and
       # "7bit") are supported.
+      #
+      # FIXME:
+      # - should use the VALUE parameter
+      # - should also take a default value type, so it can be converted
+      #   if VALUE parameter is not present.
       def value
         case encoding
-          when nil, '8bit', '7bit' then @value
+          when nil, '8BIT', '7BIT' then @value
 
           # Hack - if the base64 lines started with 2 SPC chars, which is invalid,
           # there will be extra spaces in @value. Since no SPC chars show up in
           # b64 encodings, they can be safely stripped out before unpacking.
-          when 'b', 'base64'       then @value.gsub(' ', '').unpack('m*').first
+          when 'B', 'BASE64'       then @value.gsub(' ', '').unpack('m*').first
 
-          when 'quoted-printable'  then @value.unpack('M*').first
+          when 'QUOTED-PRINTABLE'  then @value.unpack('M*').first
 
-          else raise Vpim::InvalidEncodingError, "unrecognized encoding (#{encoding})"
+          else
+            raise Vpim::InvalidEncodingError, "unrecognized encoding (#{encoding})"
         end
       end
 
       # Is the #name of this Field +name+? Names are case insensitive.
       def name?(name)
-        name.to_s.downcase == @name.downcase
+        Vpim::Methods.casecmp?(@name, name)
       end
 
       # Is the #group of this field +group+? Group names are case insensitive.
       # A +group+ of nil matches if the field has no group.
       def group?(group)
-        g1 = @group ? @group.downcase : nil
-        g2 =  group ?  group.downcase : nil
-        g1 == g2
+        Vpim::Methods.casecmp?(@group, group)
       end
 
       # Is the value of this field of type +kind+? RFC2425 allows the type of
@@ -321,7 +351,7 @@ module Vpim
       # - boolean:
       # - float:
       def kind?(kind)
-        kind.downcase == self.kind
+        Vpim::Methods.casecmp?(self.kind == kind)
       end
 
       # Is one of the values of the TYPE parameter of this field +type+? The
@@ -331,35 +361,47 @@ module Vpim
       # TYPE parameters are used for general categories, such as
       # distinguishing between an email address used at home or at work.
       def type?(type)
-        types = param('type')
+        type = type.to_str
+
+        types = param('TYPE')
 
         if types
-          types = types.include?(type.to_str.downcase)
+          types = types.detect { |t| Vpim::Methods.casecmp?(t, type) }
         end
       end
 
       # Is this field marked as preferred? A vCard field is preferred if
-      # #type?('pref'). This method is not necessarily meaningful for
+      # #type?('PREF'). This method is not necessarily meaningful for
       # non-vCard profiles.
       def pref?
-        type?('pref')
+        type? 'PREF'
+      end
+
+      # Set whether a field is marked as preferred. See #pref?
+      def pref=(ispref)
+        if ispref
+          pvalue_iadd('TYPE', 'PREF')
+        else
+          pvalue_idel('TYPE', 'PREF')
+        end
       end
 
       # Is the value of this field +value+? The check is case insensitive.
+      # FIXME - it shouldn't be insensitive, make a #casevalue? method.
       def value?(value)
-         @value && @value.downcase == value.downcase
+         Vpim::Methods.casecmp?(@value, value.to_str)
       end
 
       # The value of the ENCODING parameter, if present, or nil if not
       # present.
       def encoding
-        e = param("encoding")
+        e = param('ENCODING')
 
         if e
           if e.length > 1
-            raise Vpim::InvalidEncodingError, "multi-valued param 'encoding' (#{e})"
+            raise Vpim::InvalidEncodingError, "multi-valued param 'ENCODING' (#{e})"
           end
-          e = e.first
+          e = e.first.upcase
         end
         e
       end
@@ -367,12 +409,12 @@ module Vpim
       # The type of the value, as specified by the VALUE parameter, nil if
       # unspecified.
       def kind
-        v = param('value')
+        v = param('VALUE')
         if v
           if v.size > 1
-            raise InvalidEncodingError, "multi-valued param 'value' (#{values})"
+            raise InvalidEncodingError, "multi-valued param 'VALUE' (#{values})"
           end
-          v = v.first
+          v = v.first.downcase
         end
         v
       end
@@ -382,7 +424,7 @@ module Vpim
       # birthday). The time will be UTC if marked as so (with a timezone of
       # "Z"), and in localtime otherwise.
       #
-      # TODO: support timezone offsets
+      # TODO - support timezone offsets
       #
       # TODO - if year is before 1970, this won't work... but some people
       # are generating calendars saying Canada Day started in 1753!
@@ -440,7 +482,7 @@ module Vpim
       # characters, this method will strip them, if present.
       #
       # In theory, #value could also do this, but it would need to know that
-      # the value is of type 'text', and often for text values the 'type'
+      # the value is of type 'TEXT', and often for text values the 'VALUE'
       # parameter is not present, so knowledge of the expected type of the
       # field is required from the decoder.
       def to_text
@@ -452,10 +494,6 @@ module Vpim
         @value
       end
 
-      def inspect # :nodoc:
-        @line
-      end
-
       # TODO def pretty_print() ...
 
       # Set the group of this field to +group+.
@@ -477,21 +515,85 @@ module Vpim
       def text=(text)
       end
 
-      # Set a the param +param+'s value to +pvalue+. Valid values are as in
-      # Field.create().
-      def []=(param,pvalue)
+      # Set a the param +pname+'s value to +pvalue+, replacing any value it
+      # currently has. See Field.create() for a description of +pvalue+.
+      #
+      # Example:
+      #  if field['TYPE']
+      #    field['TYPE'] << 'HOME'
+      #  else
+      #    field['TYPE'] = [ 'HOME' ]
+      #  end
+      #
+      # TODO - this could be an alias to #pvalue_set
+      def []=(pname,pvalue)
         unless pvalue.respond_to?(:to_ary)
           pvalue = [ pvalue ]
         end
 
         h = @params.dup
 
-        h[param.downcase] = pvalue
+        h[pname.upcase] = pvalue
 
         mutate(@group, @name, h, @value)
         pvalue
       end
 
+      # Add +pvalue+ to the param +pname+'s value. The values are treated as a
+      # set so duplicate values won't occur, and String values are case
+      # insensitive.  See Field.create() for a description of +pvalue+.
+      def pvalue_iadd(pname, pvalue)
+        pname = pname.upcase
+
+        # Get a uniq set, where strings are compared case-insensitively.
+        values = [ pvalue, @params[pname] ].flatten.compact
+        values = values.collect do |v|
+          if v.respond_to? :to_str
+            v = v.to_str.upcase
+          end
+          v
+        end
+        values.uniq!
+
+        h = @params.dup
+
+        h[pname] = values
+
+        mutate(@group, @name, h, @value)
+        values
+      end
+
+      # Delete +pvalue+ from the param +pname+'s value. The values are treated
+      # as a set so duplicate values won't occur, and String values are case
+      # insensitive.  +pvalue+ must be a single String or Symbol.
+      def pvalue_idel(pname, pvalue)
+        pname = pname.upcase
+        if pvalue.respond_to? :to_str
+          pvalue = pvalue.to_str.downcase
+        end
+
+        # Get a uniq set, where strings are compared case-insensitively.
+        values = [ nil, @params[pname] ].flatten.compact
+        values = values.collect do |v|
+          if v.respond_to? :to_str
+            v = v.to_str.downcase
+          end
+          v
+        end
+        values.uniq!
+        values.delete pvalue
+
+        h = @params.dup
+
+        h[pname] = values
+
+        mutate(@group, @name, h, @value)
+        values
+      end
+
+      # FIXME - should change this so it doesn't assign to @line here, so @line
+      # is used to preserve original encoding. That way, #encode can only wrap
+      # new fields, not old fields.
       def mutate(g, n, p, v) #:nodoc:
         line = Field.encode0(g, n, p, v)
 
