@@ -1,7 +1,5 @@
 =begin
-  $Id: rfc2425.rb,v 1.10 2005/01/01 17:17:01 sam Exp $
-
-  Copyright (C) 2005 Sam Roberts
+  Copyright (C) 2008 Sam Roberts
 
   This library is free software; you can redistribute it and/or modify it
   under the same terms as the ruby language itself, see the file COPYING for
@@ -17,7 +15,8 @@ module Vpim
     # 1*(ALPHA / DIGIT / "-")
     # Note: I think I can add A-Z here, and get rid of the "i" matches elsewhere.
     # Note: added '_' to allowed because its produced by Notes - X-LOTUS-CHILD_UID
-    NAME    = '[-a-z0-9_]+'
+    # Note: added '/' to allowed because its produced by KAddressBook - X-messaging/xmpp-All:
+    NAME    = '[-a-z0-9_/]+'
 
     # <"> <Any character except CTLs, DQUOTE> <">
     QSTR    = '"([^"]*)"'
@@ -53,8 +52,19 @@ module Vpim
     # time-second = 2 DIGIT
     # time-secfrac = "," 1*DIGIT
     # time-zone = "Z" / time-numzone
-    # time-numzome = sign time-hour [":"] time-minute
+    # time-numzone = sign time-hour [":"] time-minute
     TIME = '(\d\d):?(\d\d):?(\d\d)(\.\d+)?(Z|[-+]\d\d:?\d\d)?'
+
+    # integer = (["+"] / "-") 1*DIGIT
+    INTEGER = '[-+]?\d+'
+
+    # QSAFE-CHAR = WSP / %x21 / %x23-7E / NON-US-ASCII
+    #  ; Any character except CTLs and DQUOTE
+    QSAFECHAR = '[ \t\x21\x23-\x7e\x80-\xff]'
+
+    # SAFE-CHAR  = WSP / %x21 / %x23-2B / %x2D-39 / %x3C-7E / NON-US-ASCII
+    #   ; Any character except CTLs, DQUOTE, ";", ":", ","
+    SAFECHAR = '[ \t\x21\x23-\x2b\x2d-\x39\x3c-\x7e\x80-\xff]'
   end
 end
 
@@ -62,19 +72,18 @@ module Vpim
   # Split on \r\n or \n to get the lines, unfold continued lines (they
   # start with ' ' or \t), and return the array of unfolded lines.
   #
-  # This also implements the (invalid) encoding convention of allowing empty
-  # lines to be inserted for readability - it does this by dropping
-  # zero-length lines.
+  # This also supports the (invalid) encoding convention of allowing empty
+  # lines to be inserted for readability - it does this by dropping zero-length
+  # lines.
   def Vpim.unfold(card) #:nodoc:
       unfolded = []
 
-      card.split(/\r?\n/).each do
-        |line|
-
+      card.each do |line|
+        line.chomp!
         # If it's a continuation line, add it to the last.
         # If it's an empty line, drop it from the input.
         if( line =~ /^[ \t]/ )
-            unfolded << unfolded.pop + line[1, line.size-1]
+          unfolded[-1] << line[1, line.size-1]
         elsif( line =~ /^$/ )
         else
           unfolded << line
@@ -88,21 +97,26 @@ module Vpim
   def Vpim.decode_list(value, sep = ',') # :nodoc:
     list = []
     
-    value.each(sep) {
-      |item|
-      list << yield(item) unless item =~ %r{^\s*#{sep}?$}
-    }
+    value.each(sep) do |item|
+      item.chomp!(sep)
+      list << yield(item)
+    end
     list
   end
 
   # Convert a RFC 2425 date into an array of [year, month, day].
   def Vpim.decode_date(v) # :nodoc:
-    unless v =~ %r{\s*#{Bnf::DATE}\s*}
+    unless v =~ %r{^\s*#{Bnf::DATE}\s*$}
       raise Vpim::InvalidEncodingError, "date not valid (#{v})"
     end
     [$1.to_i, $2.to_i, $3.to_i]
   end
 
+  # Convert a RFC 2425 date into a Date object.
+  def self.decode_date_to_date(v)
+    Date.new(*decode_date(v))
+  end
+
   # Note in the following the RFC2425 allows yyyy-mm-ddThh:mm:ss, but RFC2445
   # does not. I choose to encode to the subset that is valid for both.
 
@@ -123,17 +137,32 @@ module Vpim
 
   # Convert a RFC 2425 time into an array of [hour,min,sec,secfrac,timezone]
   def Vpim.decode_time(v) # :nodoc:
-    unless match = %r{\s*#{Bnf::TIME}\s*}.match(v)
-      raise Vpim::InvalidEncodingError, "time not valid (#{v})"
+    unless match = %r{^\s*#{Bnf::TIME}\s*$}.match(v)
+      raise Vpim::InvalidEncodingError, "time '#{v}' not valid"
     end
     hour, min, sec, secfrac, tz = match.to_a[1..5]
 
     [hour.to_i, min.to_i, sec.to_i, secfrac ? secfrac.to_f : 0, tz]
   end
 
-  # Convert a RFC 2425 date-time into an array of [hour,min,sec,secfrac,timezone]
+  def self.array_datetime_to_time(dtarray) #:nodoc:
+    # We get [ year, month, day, hour, min, sec, usec, tz ]
+    begin
+      tz = (dtarray.pop == "Z") ? :gm : :local
+      Time.send(tz, *dtarray)
+    rescue ArgumentError => e
+      raise Vpim::InvalidEncodingError, "#{tz} #{e} (#{dtarray.join(', ')})"
+    end
+  end
+
+  # Convert a RFC 2425 time into an array of Time objects.
+  def Vpim.decode_time_to_time(v) # :nodoc:
+    array_datetime_to_time(decode_date_time(v))
+  end
+
+  # Convert a RFC 2425 date-time into an array of [year,mon,day,hour,min,sec,secfrac,timezone]
   def Vpim.decode_date_time(v) # :nodoc:
-    unless match = %r{\s*#{Bnf::DATE}T#{Bnf::TIME}\s*}.match(v)
+    unless match = %r{^\s*#{Bnf::DATE}T#{Bnf::TIME}\s*$}.match(v)
       raise Vpim::InvalidEncodingError, "date-time '#{v}' not valid"
     end
     year, month, day, hour, min, sec, secfrac, tz = match.to_a[1..8]
@@ -146,31 +175,59 @@ module Vpim
     ]
   end
 
+  def Vpim.decode_date_time_to_datetime(v) #:nodoc:
+    year, month, day, hour, min, sec, secfrac, tz = Vpim.decode_date_time(v)
+    # TODO - DateTime understands timezones, so we could decode tz and use it.
+    DateTime.civil(year, month, day, hour, min, sec, 0)
+  end
+
   # Vpim.decode_boolean
   #
   # float
   #
   # float_list
-  #
-  # integer
+=begin
+=end
+
+  # Convert an RFC2425 INTEGER value into an Integer
+  def Vpim.decode_integer(v) # :nodoc:
+    unless match = %r{\s*#{Bnf::INTEGER}\s*}.match(v)
+      raise Vpim::InvalidEncodingError, "integer not valid (#{v})"
+    end
+    v.to_i
+  end
+
   #
   # integer_list
-  #
-  # text_list
 
-  # Convert a RFC 2425 date-list into an array of dates.
+  # Convert a RFC2425 date-list into an array of dates.
   def Vpim.decode_date_list(v) # :nodoc:
-    dates = Vpim.decode_list(v) { |date| Vpim.decode_date(date) }
+    Vpim.decode_list(v) do |date|
+      date.strip!
+      if date.length > 0
+        Vpim.decode_date(date)
+      end
+    end.compact
   end
 
   # Convert a RFC 2425 time-list into an array of times.
   def Vpim.decode_time_list(v) # :nodoc:
-    times = Vpim.decode_list(v) { |time| Vpim.decode_time(time) }
+    Vpim.decode_list(v) do |time|
+      time.strip!
+      if time.length > 0
+        Vpim.decode_time(time)
+      end
+    end.compact
   end
 
   # Convert a RFC 2425 date-time-list into an array of date-times.
   def Vpim.decode_date_time_list(v) # :nodoc:
-    datetimes = Vpim.decode_list(v) { |datetime| Vpim.decode_date_time(datetime) }
+    Vpim.decode_list(v) do |datetime|
+      datetime.strip!
+      if datetime.length > 0
+        Vpim.decode_date_time(datetime)
+      end
+    end.compact
   end
 
   # Convert RFC 2425 text into a String.
@@ -178,8 +235,69 @@ module Vpim
   # \n -> NL
   # \N -> NL
   # \, -> ,
+  # \; -> ;
+  #
+  # I've seen double-quote escaped by iCal.app. Hmm. Ok, if you aren't supposed
+  # to escape anything but the above, everything else is ambiguous, so I'll
+  # just support it.
   def Vpim.decode_text(v) # :nodoc:
-    v.gsub(/\\[nN]/, "\n").gsub(/\\,/, ",").gsub(/\\\\/) { |m| "\\" }
+    # FIXME - I think this should trim leading and trailing space
+    v.gsub(/\\(.)/) do
+      case $1
+      when 'n', 'N' 
+        "\n"
+      else
+        $1
+      end
+    end
+  end
+  
+  def Vpim.encode_text(v) #:nodoc:
+    v.to_str.gsub(/([\\,;\n])/) { $1 == "\n" ? "\\n" : "\\"+$1 }
+  end
+
+  # v is an Array of String, or just a single String
+  def Vpim.encode_text_list(v, sep = ",") #:nodoc:
+    begin
+      v.to_ary.map{ |t| Vpim.encode_text(t) }.join(sep)
+    rescue
+      Vpim.encode_text(v)
+    end
+  end
+
+  # Convert a +sep+-seperated list of TEXT values into an array of values.
+  def Vpim.decode_text_list(value, sep = ',') # :nodoc:
+    # Need to do in two stages, as best I can find.
+    list = value.scan(/([^#{sep}\\]*(?:\\.[^#{sep}\\]*)*)#{sep}/).map do |v|
+      Vpim.decode_text(v.first)
+    end
+    if value.match(/([^#{sep}\\]*(?:\\.[^#{sep}\\]*)*)$/)
+      list << $1
+    end
+    list
+  end
+
+  # param-value = paramtext / quoted-string
+  # paramtext  = *SAFE-CHAR
+  # quoted-string      = DQUOTE *QSAFE-CHAR DQUOTE
+  def Vpim.encode_paramtext(value)
+    case value
+    when %r{\A#{Bnf::SAFECHAR}*\z}
+      value
+    else
+      raise Vpim::Unencodable, "paramtext #{value.inspect}"
+    end
+  end
+
+  def Vpim.encode_paramvalue(value)
+    case value
+    when %r{\A#{Bnf::SAFECHAR}*\z}
+      value
+    when %r{\A#{Bnf::QSAFECHAR}*\z}
+      '"' + value + '"'
+    else
+      raise Vpim::Unencodable, "param-value #{value.inspect}"
+    end
   end
 
 
@@ -203,13 +321,13 @@ module Vpim
     current = [ dst ]
 
     for f in src
-      if f.name? 'begin'
+      if f.name? 'BEGIN'
         e = [ f ]
 
         current.last.push(e)
         current.push(e)
 
-      elsif f.name? 'end'
+      elsif f.name? 'END'
         current.last.push(f)
 
         unless current.last.first.value? current.last.last.value
@@ -230,6 +348,7 @@ module Vpim
   # an array of all the inner arrays of fields. Return the array [outer,
   # inner].
   def Vpim.outer_inner(fields) #:nodoc:
+    # FIXME - use Enumerable#partition
     # seperate into the outer-level fields, and the arrays of component
     # fields
     outer = []
