@@ -1,7 +1,5 @@
 =begin
-  $Id: rrule.rb,v 1.18 2005/01/23 22:19:48 sam Exp $
-
-  Copyright (C) 2005 Sam Roberts
+  Copyright (C) 2008 Sam Roberts
 
   This library is free software; you can redistribute it and/or modify it
   under the same terms as the ruby language itself, see the file COPYING for
@@ -18,6 +16,12 @@ require 'pp'
 
 $debug = ENV['DEBUG']
 
+class Date
+  def inspect
+    self.to_s
+  end
+end
+
 def debug(*objs)
   if $debug
     pp(*objs)
@@ -28,21 +32,18 @@ end
 
 module Vpim
 
-  # Implements the iCalendar recurence rule syntax. See etc/rrule.txt for the
+  # Implements the iCalendar recurrence rule syntax. See etc/rrule.txt for the
   # syntax description and examples from RFC 2445. The description is pretty
   # hard to understand, but the examples are more helpful.
   #
-  # The implementation is pretty complete, but still lacks support for:
+  # The implementation is reasonably complete, but still lacks support for:
   #
-  # TODO - BYWEEKLY, BYWEEKNO, WKST: rules that recur by the week, or are
-  # limited to particular weeks, not hard, but not trivial, I'll do it for the
-  # next release
+  # Recurrence by date (RDATE) and exclusions (EXDATE, EXRULE).
+  #
+  # TODO - BYWEEKNO: rules that are limited to particular weeks in a year.
   #
   # TODO - BYHOUR, BYMINUTE, BYSECOND: trivial to do, but I don't have an
-  # immediate need for them, I'll do it for the next release
-  #
-  # TODO - BYSETPOS: limiting to only certain recurrences in a set (what does
-  # -1, last occurence, mean for an infinitely occuring rule?)
+  # immediate need for them.
   #
   # TODO - new API? -> Rrule#infinite?
   #
@@ -86,7 +87,7 @@ module Vpim
 
            case key
            when 'FREQ'
-             @freq = value.upcase
+             @freq = value
 
            when 'UNTIL'
              if @count
@@ -123,7 +124,7 @@ module Vpim
        end
     end
 
-    # Return an Enumerable, it's #each() will yield over all occurences up to
+    # Return an Enumerable, it's #each() will yield over all occurrences up to
     # (and not including) time +dountil+.
     def each_until(dountil)
       Vpim::Enumerator.new(self, dountil)
@@ -132,25 +133,27 @@ module Vpim
     # Yields for each +ytime+ in the recurring set of events.
     #
     # Warning: the set may be infinite! If you need an upper bound on the
-    # number of occurences, you need to implement a count, or pass a time,
+    # number of occurrences, you need to implement a count, or pass a time,
     # +dountil+, which will not be iterated past (i.e. all times yielded will be
     # less than +dountil+).
     #
     # Also, iteration will not currently continue past the limit of a Time
-    # object, which is some time in 2037 with the 32-bit time_t's common on
+    # object, which is some time in 2037 with the 32-bit time_t common on
     # most systems.
     def each(dountil = nil) #:yield: ytime
       t = @dtstart.clone
-      count = 1
 
       # Time.to_a => [ sec, min, hour, day, month, year, wday, yday, isdst, zone ]
 
-      # Every event occurs at least once, at its start time, but only if the start
-      # time is earlier than 'dountil'...
+      # Every event occurs at its start time, but only if the start time is
+      # earlier than DOUNTIL...
+      if !dountil || t < dountil
+        yield t
+      end
+      count = 1
+
+      # With no recurrence, DTSTART is the only occurrence.
       if !@rrule
-        if !dountil || t < dountil
-          yield t
-        end
         return self
       end
 
@@ -168,24 +171,34 @@ module Vpim
 
         case @freq
           #when 'YEARLY' then
-          # Don't need to keep track of year, all occurences are within t's
+          # Don't need to keep track of year, all occurrences are within t's
           # year.
-          when 'MONTHLY'  then  days.month = t.month #month = { t.month => nil }
-  #       when 'WEEKLY'   then  days.mday = t.month, t.mday
+        when 'MONTHLY'  then  days.month = t.month
+        when 'WEEKLY'   then  #days.month = t.month
           # TODO - WEEKLY
-          when 'DAILY'    then  days.mday = t.month, t.mday #month = { t.month => [ t.mday ] }
-          when 'HOURLY'   then  hour  = [t.hour]
-          when 'MINUTELY' then  min   = [t.min]
-          when 'SECONDLY' then  sec   = [t.sec]
+        when 'DAILY'    then  days.mday = t.month, t.mday
+        when 'HOURLY'   then  hour  = [t.hour]
+        when 'MINUTELY' then  min   = [t.min]
+        when 'SECONDLY' then  sec   = [t.sec]
         end
 
+  #      debug [t, days]
         # Process the BY* modifiers in RFC defined order:
-        #  BYMONTH, BYWEEKNO, BYYEARDAY, BYMONTHDAY, BYDAY,
-        #  BYHOUR, BYMINUTE, BYSECOND and BYSETPOS
-        
+        #  BYMONTH,
+        #  BYWEEKNO,
+        #  BYYEARDAY,
+        #  BYMONTHDAY,
+        #  BYDAY,
+        #  BYHOUR,
+        #  BYMINUTE,
+        #  BYSECOND,
+        #  BYSETPOS
+
+        bymon = [nil]
+
         if @by['BYMONTH']
           bymon = @by['BYMONTH'].split(',')
-          bymon = bymon.collect { |m| m.to_i }
+          bymon = bymon.map { |m| m.to_i }
   #        debug bymon
 
           # In yearly, at  this point, month will always be nil. At other
@@ -220,21 +233,20 @@ module Vpim
 
         if @by['BYDAY']
           byday = @by['BYDAY'].scan(/,?([+-]?[1-9]?\d*)?(SU|MO|TU|WE|TH|FR|SA)/i)
-  #        debug byday
 
           # BYDAY means different things in different frequencies. The +n+
           # is only meaningful when freq is yearly or monthly.
 
           case @freq
             when 'YEARLY'
-              dates = byday_in_yearly(t.year, byday)
+              dates = bymon.map { |m| byday_in_monthly(t.year, m, byday) }.flatten
             when 'MONTHLY'
               dates = byday_in_monthly(t.year, t.month, byday)
             when 'WEEKLY'
-              # dates = byday_in_weekly(t.year, wkstart, t.month, t.day, byday)
+              dates = byday_in_weekly(t.year, t.month, t.mday, @wkst, byday)
             when 'DAILY', 'HOURLY', 'MINUTELY', 'SECONDLY'
-              # Reuse the bday_in_monthly. Current day is already specified,
-              # so this will just eliminate the current day if its no allowed
+              # Reuse the byday_in_monthly. Current day is already specified,
+              # so this will just eliminate the current day if its not allowed
               # in BYDAY.
               dates = byday_in_monthly(t.year, t.month, byday)
           end
@@ -246,48 +258,73 @@ module Vpim
 
         # TODO - BYHOUR, BYMINUTE, BYSECOND
         
-        # TODO - BYSETPOS
-
-        # Yield the time, if we haven't gone over COUNT, or past UNTIL, or past
-        # the end of representable time.
-
         hour   = [@dtstart.hour]   if !hour 
         min    = [@dtstart.min]    if !min  
         sec    = [@dtstart.sec]    if !sec 
 
   #      debug days
 
+        # Generate the yield set so BYSETPOS can be evaluated.
+        yset = []
+
         days.each do |m,d|
-            hour.each do |h|
-              min.each do |n|
-                sec.each do |s|
-                  if(@count && (count > @count))
-                    return self
-                  end
-                  y = Time.local(t.year, m, d, h, n, s, 0)
+          hour.each do |h|
+            min.each do |n|
+              sec.each do |s|
+                y = Time.local(t.year, m, d, h, n, s, 0)
 
-                  next if y.hour != h
+                next if y.hour != h
 
-                  # The generated set can sometimes generate results earlier
-                  # than the DTSTART, skip them.
-                  next if y < @dtstart
-
-                  # We are done if current time is past @until.
-                  if @until && (y > @until)
-                    return self
-                  end
-                  # We are also done if current time is past the
-                  # caller-requested until.
-                  if dountil && (y > dountil)
-                    return self
-                  end
-                  yield y
-                  count += 1
-                end
+                yset << y
               end
             end
+          end
         end
 
+        if @by['BYSETPOS']
+          bysetpos = @by['BYSETPOS'].split(',')
+          yset = bysetpos.map do |i|
+            i = i.to_i
+            case
+            when i < 0
+              # yset[-1] is last
+              yset[i]
+            when i > 0
+              # yset[1] is first
+              yset[i-1]
+            else
+              # ignore invalid syntax
+            end
+          end.compact # set positions out of scope will be nil, RFC says ignore them
+        end
+
+        # Yield the occurrence, if we haven't gone over COUNT, or past UNTIL, or
+        # past the end of representable time.
+
+        yset.each do |y|
+          # The generated set can sometimes generate results earlier
+          # than the DTSTART, skip them. Also, we already yielded
+          # DTSTART, skip it.
+          next if y <= @dtstart
+
+          count += 1
+
+          # We are done if current count is past @count.
+          if(@count && (count > @count))
+            return self
+          end
+
+          # We are done if current time is past @until.
+          if @until && (y > @until)
+            return self
+          end
+          # We are also done if current time is past the
+          # caller-requested until.
+          if dountil && (y >= dountil)
+            return self
+          end
+          yield y
+        end
 
         # Add @interval to @freq component
 
@@ -330,6 +367,8 @@ module Vpim
       end
     end
 
+
+
     class DaySet #:nodoc:
 
       def initialize(ref)
@@ -363,30 +402,30 @@ module Vpim
       end
 
       def intersect_dates(dates) #:nodoc:
-        if dates
-          # If no months are in the dayset, add all the ones in dates
-          if !@month
-            @month = {}
+        return unless dates
 
-            dates.each do |d|
-              @month[d.mon] = nil
-            end
+        # If no months are in the dayset, add all the ones in dates
+        if !@month
+          @month = {}
+
+          dates.each do |d|
+            @month[d.mon] = nil
           end
+        end
 
-          # In each month,
-          #   if there are days,
-          #     eliminate those not in dates
-          #   otherwise
-          #     add all those in dates
-          @month.each do |mon, days|
-            days_in_mon = dates.find_all { |d| d.mon == mon }
-            days_in_mon = days_in_mon.collect { |d| d.day }
+        # In each month,
+        #   if there are days,
+        #     eliminate those not in dates
+        #   otherwise
+        #     add all those in dates
+        @month.each do |mon, days|
+          days_in_mon = dates.find_all { |d| d.mon == mon }
+          days_in_mon = days_in_mon.map { |d| d.day }
 
-            if days
-              days_in_mon = days_in_mon & days
-            end
-            @month[mon] = days_in_mon
+          if days
+            days_in_mon = days_in_mon & days
           end
+          @month[mon] = days_in_mon
         end
       end
 
@@ -451,16 +490,12 @@ module Vpim
       dates = []
 
       byyday.each do |yday|
-        dates << Date.new2(year, yday[0].to_i)
+        dates << Date.ordinal(year, yday[0].to_i)
       end
       dates.sort!
       dates
     end
 
-    def byday_in_yearly(year, byday) #:nodoc:
-      byday_in_monthly(year, nil, byday)
-    end
-
     def byday_in_monthly(year, mon, byday) #:nodoc:
       dates = []
 
@@ -476,6 +511,17 @@ module Vpim
       dates
     end
 
+    def byday_in_weekly(year, mon, day, wkst, byday)
+      #    debug ["day", year,mon,day,wkst,byday]
+      days = byday.map{ |_, byday| Date.str2wday(byday) }
+      week = DateGen.weekofdate(year, mon, day, wkst)
+      #    debug [ "week", dates ]
+      week.delete_if do |d|
+        !days.include?(d.wday)
+      end
+      week
+    end
+
   end
 
 end
