@@ -1,21 +1,25 @@
-#!/usr/local/bin/ruby18 -w
-# $Id: reminder.rb,v 1.7 2004/11/21 18:30:13 sam Exp $
+#!/usr/bin/env ruby
 
-$:.unshift File.dirname($0)
+$-w = true
+
+require 'ubygems' rescue "ignored"
 
 require 'getoptlong'
 require 'pp'
+require 'plist'
 
-require 'vpim/icalendar'
-require 'vpim/duration'
+require 'vpim/repo'
+
+$stdout.sync = true
+$stderr.sync = true
 
 HELP =<<EOF
-Usage: #{$0} <calendar>...
+Usage: #{$0} [where]
 
 Shows events and todos occuring soon.
 
-By default, the calendar files from ~/Library/Calendars are used (the
-location of Apple's iCal calendars).
+By default, the Apple iCal v3 calendars are used, but if a location where
+.ics files is specified, any calendars found there will be used.
 
 Options
   -h,--help        Print this helpful message.
@@ -53,10 +57,16 @@ opts.each do |opt, arg|
   end
 end
 
+calendars = []
+
 if ARGV.length > 0
-  calendars = ARGV
+  Vpim::Repo::Directory.each(ARGV.first) do |cal|
+    calendars << cal
+  end
 else
-  calendars = Dir[ File.expand_path("~/Library/Calendars/*.ics") ]
+  Vpim::Repo::Ical3.each() do |cal|
+    calendars << cal
+  end
 end
 
 if opt_debug
@@ -81,37 +91,34 @@ if opt_verbose
 end
 
 # Collect all events, then all todos.
-allEvents = []
-allTodos  = []
+all_events = []
+all_todos  = []
 
-def Fu(a, b)
-  pp a
-  pp b
-end
+calendars.each do |cal|
+  if opt_debug; puts cal.name; end
 
-calendars.each do |file|
-  if opt_debug; puts file; end
-
-  next if File.basename(file) =~ /^x-/
-
-  cals = Vpim::Icalendar.decode(File.open(file))
-
-  cals.each do |cal|
+  begin
     cal.events.each do |e|
-      if opt_debug; pp e; end
-      if e.occurs_in?(t0, t1)
-        if e.summary
-          allEvents.push(e)
+      begin
+        if opt_debug; pp e; end
+        if e.occurs_in?(t0, t1)
+          if e.summary
+            all_events.push(e)
+          end
         end
+      rescue
+        $stderr.puts "error in #{cal.name} (\"#{e.summary}\"): #{$!.to_s}"
       end
     end
 
-    allTodos.concat(cal.todos)
+    all_todos.concat(cal.todos)
   end
 end
 
-def StartOfFirstOccurence(t0, t1, e)
-  e.occurences.each_until(t1).each do |t|
+puts
+
+def start_of_first_occurrence(t0, t1, e)
+  e.occurrences.each_until(t1).each do |t|
     # An event might start before t0, but end after it..., in which case
     # we are still interested.
     if (t + (e.duration || 0)) >= t0
@@ -121,25 +128,24 @@ def StartOfFirstOccurence(t0, t1, e)
   nil
 end
 
-allEvents.sort! do |lhs, rhs|
-  StartOfFirstOccurence(t0, t1, lhs) <=> StartOfFirstOccurence(t0, t1, rhs)
+all_events.sort! do |lhs, rhs|
+  start_of_first_occurrence(t0, t1, lhs) <=> start_of_first_occurrence(t0, t1, rhs)
 end
 
-allEvents.each do |e|
+all_events.each do |e|
   puts "#{e.summary}:"
 
   if opt_verbose
     if e.description;   puts "  description=#{e.description}"; end
-    if e.comment;       puts "  comment=#{e.comment}"; end
+    if e.comments;      puts "  comment=#{e.comments.first}"; end
     if e.location;      puts "  location=#{e.location}"; end
     if e.status;        puts "  status=#{e.status}"; end
-    if e.rrule;         puts "  rrule=#{e.rrule}"; end
     if e.dtstart;       puts "  dtstart=#{e.dtstart}"; end
     if e.duration;      puts "  duration=#{Vpim::Duration.new(e.duration).to_s}"; end
   end
 
   i = 1
-  e.occurences.each_until(t1).each do |t|
+  e.occurrences.each_until(t1).each do |t|
     # An event might start before t0, but end after it..., in which case
     # we are still interested.
     dstr = ''
@@ -164,7 +170,7 @@ def fix_priority(vtodo)
 end
 =end
 
-allTodos.sort! do |x,y|
+all_todos.sort! do |x,y|
   x = x.priority
   y = y.priority
 
@@ -188,7 +194,7 @@ priorities = [
   'not important'
 ]
 
-allTodos.each do |e|
+all_todos.each do |e|
   status = e.status || 'Todo'
   if status != 'COMPLETED'
     puts "#{status.capitalize}: #{e.summary}" #  (#{priorities[e.priority]})"
